# coding: utf-8
# Copyright (c) Qotto, 2021

"""Eventy protocol data objects

Define a base Record class, and the 3 event types defined by the eventy protocol.
"""

from .event import Event
from .record import Record
from .record_errors import RecordAttributeTypeError, RecordAttributeValueError, RecordAttributeError
from .request import Request
from .response import Response

__all__ = [
    'Record',
    'Event',
    'Request',
    'Response',
    'RecordAttributeError',
    'RecordAttributeValueError',
    'RecordAttributeTypeError',
]



# coding: utf-8
# Copyright (c) Qotto, 2021

from .context_filters import add_correlation_id_filter, add_request_id_filter
from .gke_handler import GkeHandler
from .simple_handler import SimpleHandler

__all__ = [
    'add_correlation_id_filter',
    'add_request_id_filter',
    'SimpleHandler',
    'GkeHandler',
]



# coding: utf-8
# Copyright (c) Qotto, 2021

import logging
from base64 import b64encode
from datetime import datetime, timezone
from secrets import token_urlsafe
from typing import Callable

from ..context import EventyContext

logger = logging.getLogger(__name__)

__all__ = [
    'with_context',
    'merge_decorators',
    'DecoratorType',
    'generate_simple_id',
    'simple_func_id_generator',
]

DecoratorType = Callable[..., Callable]


def merge_decorators(*decorators) -> DecoratorType:
    """
    Merge multiple decorators, apply from left to right
    """

    if not decorators:
        def identity(func):
            return func

        return identity

    def decorator(func):
        def decorated_func(*args, **kwargs):
            inner_func = func
            nested_decorators = decorators
            for nested_decorator in nested_decorators:
                inner_func = nested_decorator(inner_func)

            return inner_func(*args, **kwargs)

        return decorated_func

    return decorator


def with_context(
    context: EventyContext,
    generator: Callable[[Callable, tuple, dict], str],
    reset: bool = False,
    fetch: bool = True,
) -> DecoratorType:
    def decorator(func):
        def decorated_func(*args, **kwargs):
            token = None
            if context.name in kwargs and fetch:
                token = context.set(kwargs.pop(context.name))
                logger.debug(f'Extracted {context.name} from {func.__name__} kwargs')
            elif generator:
                token = context.set(generator(func, *args, **kwargs))
                logger.debug(f'Generated {context.name} for {func.__name__}')
            else:
                logger.debug(f'No {context.name} for {func.__name__}')

            result = func(*args, **kwargs)

            if token and reset:
                logger.debug(f'Reset {context.name} for {func.__name__}')
                context.reset(token)

            return result

        return decorated_func

    return decorator


def generate_simple_id(prefix, content):
    """
    Create a context is string with "prefix:content:encoded_date:random"
    """
    ts_now = datetime.now(timezone.utc).timestamp()
    ts_2k = datetime(2000, 1, 1, tzinfo=timezone.utc).timestamp()
    ts2000res65536 = int(65536 * (ts_now - ts_2k)).to_bytes(6, 'big')
    date = b64encode(ts2000res65536, b'_-').decode('ascii')
    random = token_urlsafe(3)
    id = f'{prefix}:{content}:{date}:{random}'
    logger.debug(f'Generated context id: {id}')
    return id


def simple_func_id_generator(namespace: str = 'default_namespace'):
    def id_generator(func: Callable, *args, **kwargs):
        return generate_simple_id(namespace, func.__name__)

    return id_generator



# coding: utf-8
# Copyright (c) Qotto, 2021

import json
import logging
import math

from .context_filters import add_correlation_id_filter, add_request_id_filter
from ..context import correlation_id_ctx, request_id_ctx


class GkeHandler(logging.StreamHandler):
    def __init__(self, stream=None, level='DEBUG'):
        super().__init__(stream)
        self.addFilter(add_correlation_id_filter)
        self.addFilter(add_request_id_filter)
        self.setLevel(level=level)

    def format(self, record: logging.LogRecord) -> str:
        message = super().format(record)

        subsecond, second = math.modf(record.created)
        payload = {
            'timestamp': {'seconds': int(second), 'nanos': int(subsecond * 1e9)},
            'severity': record.levelname,
            'message': message,
            'file': record.pathname,
            'line': record.lineno,
            'module': record.module,
            'function': record.funcName,
            'logger_name': record.name,
            'thread': record.thread,
        }
        if correlation_id_ctx.is_defined():
            payload['correlation_id'] = correlation_id_ctx.get()
        if request_id_ctx.is_defined():
            payload['request_id'] = request_id_ctx.get()
        return json.dumps(payload)



# coding: utf-8
# Copyright (c) Qotto, 2021

from contextvars import ContextVar, Token


class EventyContext:
    _contextvar: ContextVar

    def __init__(self, var_name: str):
        self._contextvar = ContextVar(var_name, default="")

    @property
    def name(self):
        return self._contextvar.name

    def get(self) -> str:
        return self._contextvar.get()

    def set(self, value: str) -> Token:
        return self._contextvar.set(value)

    def is_defined(self) -> bool:
        return self.get() != ''

    def unset(self) -> None:
        self._contextvar.set('')

    def reset(self, token: Token) -> None:
        self._contextvar.reset(token)


correlation_id_ctx = EventyContext("correlation_id")
request_id_ctx = EventyContext("request_id")



# coding: utf-8
# Copyright (c) Qotto, 2021

from abc import ABC

from .base import Consumer
from ..record import Record


class TransactionalProducer(ABC):
    """
    Abstract base class for a transactional record producer
    """

    def add_to_transaction(self, destination: str, record: Record, *arg, **kwargs) -> None:
        """
        Add a record to be committed later

        :param destination: Destination of the record
        :param record: Record to be sent
        """
        raise NotImplementedError

    def commit(self, *arg, **kwargs):
        """Produce messages added to transaction in an atomic operation"""
        raise NotImplementedError


class TransactionalProcessor(Consumer, TransactionalProducer, ABC):
    """Produce records and commit read atomically"""

    def commit(self, *arg, **kwargs):
        """Produce messages added to transaction and commit read in an atomic operation"""
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2021

from typing import Union

from .. import RecordSerializer
from ...record import Record, Event, Request, Response


class AvroSerializer(RecordSerializer):

    def encode(self, record: Record) -> bytes:
        raise NotImplementedError

    def decode(self, encoded: bytes) -> Union[Event, Request, Response]:
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2021

from typing import Any, Dict

from ..record import Record


class RecordSerializer:
    """Record serializer interface"""

    def encode(self, record: Record) -> Any:
        """Encode a record

        Raises:
            SerializationError
        """
        raise NotImplementedError

    def decode(self, encoded: Any) -> Record:
        """Decode a record

        Raises:
            SerializationError
        """
        raise NotImplementedError


class ContextRecordSerializer:
    def encode(self, context: Any, record: Record) -> Any:
        raise NotImplementedError

    def decode(self, context: Any, encoded: Any) -> Record:
        raise NotImplementedError


class DestinationRecordSerializer(ContextRecordSerializer):
    def __init__(
        self,
        default_serializer: RecordSerializer,
        destination_serializers: Dict[str, RecordSerializer] = None,
    ):
        self._default_serializer = default_serializer
        if not destination_serializers:
            destination_serializers = {}
        self._destination_serializers = destination_serializers.copy()

    def _select_serializer(self, context) -> RecordSerializer:
        serializer = self._destination_serializers.get(context)
        if serializer:
            return serializer
        else:
            return self._default_serializer

    def encode(self, context: str, record: Record) -> Any:
        serializer = self._select_serializer(context)
        return serializer.encode(record)

    def decode(self, context: str, encoded: Any) -> Record:
        serializer = self._select_serializer(context)
        return serializer.decode(encoded)



# coding: utf-8
# Copyright (c) Qotto, 2021

import logging
from typing import Callable, Any

from .common import DecoratorType, merge_decorators
from ..context import correlation_id_ctx

logger = logging.getLogger(__name__)

try:
    from celery.result import AsyncResult

    logger.debug("Celery support enabled")
except ImportError:
    AsyncResult = Any
    logger.debug("Celery support disabled")

__all__ = [
    'traced',
    'namespaced',
]


def traced(celery_send_task: Callable) -> Callable[[str, list, dict, ], AsyncResult]:
    """
    Wraps a celery send_task method, adding context to task_kwargs
    """

    def traced_send_task(task_name: str, task_args: list, task_kwargs: dict, *args, **kwargs) -> AsyncResult:
        """
        Modified version of celery.execute.send_task adding context in task kwargs
        """
        if correlation_id_ctx.is_defined():
            task_kwargs.update(trace_id=correlation_id_ctx.get())
            logger.debug('Added correlation_id in task kwargs')
        else:
            logger.debug('No correlation_id in context to add in task kwargs')

        return celery_send_task(task_name, task_args, task_kwargs, *args, **kwargs)

    return traced_send_task


def namespaced(namespace: str, celery_decorator, *other_decorators) -> DecoratorType:
    """
    Wraps celery task and share_task decorators to include namespace in task name
    """

    def auto_decorator(*decorator_args, **decorator_kwargs) -> DecoratorType:
        logger.info(f'auto_decorator decorator_args={decorator_args} - decorator_kwargs={decorator_kwargs}')

        def decorator(func: Callable):
            def decorated_func(*func_args, **func_kwargs):
                full_name = f'{namespace}.tasks.{func.__name__}'
                decorator_kwargs.update(name=full_name)
                inner_func = merge_decorators(*other_decorators)(func)
                return celery_decorator(**decorator_kwargs)(inner_func)(*func_args, **func_kwargs)

            return decorated_func

        if len(decorator_args) == 1 and callable(decorator_args[0]):
            # @decorator
            # def f():
            return decorator(decorator_args[0])
        else:
            # @decorator(param=value)
            # def f():
            return decorator

    return auto_decorator



# coding: utf-8
# Copyright (c) Qotto, 2021

from .eventy_context import correlation_id_ctx, request_id_ctx, EventyContext

__all__ = [
    'correlation_id_ctx',
    'request_id_ctx',
    'EventyContext',
]



# coding: utf-8
# Copyright (c) Qotto, 2021
from abc import ABC
from typing import Iterable

import logging
from ..record import Record

logger = logging.getLogger(__name__)


class Consumer(ABC):
    """
    Abstract base class for a record consumer

    Allows batch polling, and manual or automatic after processing commit
    """

    def poll_commit(self, nb: int = 1, timeout: float = 0, *args, **kwargs) -> Iterable[Record]:
        """
        Infinite generator methods, with auto commit

        Poll records as batches of `nb` records, and commit the batch after the last record has been yielded.
        Previously polled records can be committed as soon as we are waiting for the next one.

        If records need to be polled ahead of time (e.g. for parallelization) and not be committed until after
        they are processed, the `poll()` method should be used.

        In case of synchronous processing, this method guarantees that messages are committed only after processing
        and the number of messages polled and not committed is at most `nb`.

        :param nb: Batch size, default 1
        :param timeout: Timeout for the batch poll in seconds, default 0 for infinite timeout
        :return: Polled records iterable
        :raise ConsumerPollError: On polling failure
        :raise ConsumerCommitError: On commit failure
        """
        while True:
            n = 0
            for record in self.poll(nb, timeout, *args, **kwargs):
                yield record
                n += 1
            if n > 0:
                logger.debug(f'Consumer will commit {n} records')
                self.commit()

    def poll(self, nb: int = 1, timeout: float = 0, *args, **kwargs) -> Iterable[Record]:
        """
        Poll a batch of records, without committing

        :param nb: Batch size, default 1
        :param timeout: Timeout for the batch poll in seconds, default 0 for infinite timeout
        :return: Polled records iterable
        :raise ConsumerPollError:
     """
        raise NotImplementedError

    def commit(self, *arg, **kwargs) -> None:
        """
        Commit all previously polled records

        Synchronous method, all previously polled records are committed on return

        :raise ConsumerCommitError:
        """
        raise NotImplementedError


class Producer(ABC):
    """
    Abstract base class for a record producer

    Allows to produce records synchronously
    """

    def produce_now(self, destination: str, record: Record, *arg, **kwargs) -> None:
        """
        Produce immediately a single record

        Synchronous method, the record is sent on return.

        :param destination: Destination of the record (e.g. a kafka topic)
        :param record: The record to be produced

        :raise ProducerProduceError: The record could not be produced
        """
        raise NotImplementedError



# coding: utf-8
# Copyright (c) Qotto, 2021

from logging import Formatter, StreamHandler

from coloredlogs import ColoredFormatter

from .context_filters import add_correlation_id_filter, add_request_id_filter


class SimpleHandler(StreamHandler):
    def __init__(
        self,
        fmt='%(asctime)s [CID:%(correlation_id)s] [RID:%(request_id)s] '
            '%(levelname)s %(name)s (%(module)s L%(lineno)d) '
            '%(message)s',
        datefmt='[%Y-%m-%d %H:%M:%S %z]',
        colored=False,
        stream=None,
        level='DEBUG',
    ):
        super().__init__(stream)
        self.addFilter(add_correlation_id_filter)
        self.addFilter(add_request_id_filter)
        if colored:
            self.setFormatter(ColoredFormatter(fmt=fmt, datefmt=datefmt))
        else:
            self.setFormatter(Formatter(fmt=fmt, datefmt=datefmt))
        self.setLevel(level=level)



# coding: utf-8
# Copyright (c) Qotto, 2021

from datetime import datetime
from typing import Dict, Any

from urnparse import URN8141, InvalidURNFormatError

from .record import Record
from .record_errors import RecordAttributeValueError


class Request(Record):

    def __init__(
        self,
        protocol_version: str,
        schema: str,
        version: str,
        source: str,
        destination: str,
        uuid: str = None,
        correlation_id: str = None,
        partition_key: str = None,
        date: datetime = None,
        context: Dict[str, Dict[str, Any]] = None,
        data: Dict[str, Any] = None

    ):
        """
        Initialize Record attributes

        :param protocol_version: SemVer version of the Eventy protocol
        :param schema: URN of the record schema
        :param version: SemVer version of the record schema
        :param source: URN of the record emitter
        :param destination: URN of the request destination
        :param uuid: UUID v4 of the record. Default: auto-generated
        :param correlation_id: Trace / correlation ID for trace_id. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Keys should be URN of a service using the context
        :param data: Domain data

        :raises RecordAttributeError:
        """
        super().__init__(
            protocol_version,
            schema,
            version,
            source,
            uuid,
            correlation_id,
            partition_key,
            date,
            context,
            data,
        )
        self.destination = destination

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, destination: str):
        try:
            URN8141.from_string(destination)
        except InvalidURNFormatError:
            raise RecordAttributeValueError('destination', destination, 'Not in URN format')
        self._destination = destination

    @property
    def type(self) -> str:
        return 'REQUEST'



# coding: utf-8
# Copyright (c) Qotto, 2021

class SerializationError(Exception):
    pass


class UnknownRecordTypeError(SerializationError):
    def __init__(self, record_type: str):
        super().__init__(f'Unknown record type {record_type}')



# coding: utf-8
# Copyright (c) Qotto, 2021

"""Kafka client implementation of the Eventy messaging API

This implementation uses the Confluent Python Client, which is a wrapper around librdkafka
"""

from .kafka_admin import KafkaAdmin
from .kafka_base import KafkaConsumer, KafkaProducer
from .kafka_transactional import KafkaTransactionalProcessor, KafkaTransactionalProducer

__all__ = [
    'KafkaConsumer',
    'KafkaProducer',
    'KafkaTransactionalProcessor',
    'KafkaTransactionalProducer',
    'KafkaAdmin',
]



# coding: utf-8
# Copyright (c) Qotto, 2021

from .record import Record


class Event(Record):
    """
    Event implementation of the Record abstract base class
    """

    @property
    def type(self):
        return 'EVENT'



# coding: utf-8
# Copyright (c) Qotto, 2021

class MessagingError(BaseException):
    """
    Base Error class of all messaging API
    """
    pass


class ConsumerCreationError(MessagingError):
    """
    A consumer could not be initialized
    """
    pass


class ConsumerPollError(MessagingError):
    """
    A consumer could not poll messages
    """
    pass


class ConsumerCommitError(MessagingError):
    """
    A consumer could not commit messages
    """
    pass


class ProducerCreationError(MessagingError):
    """
    A producer could not be initialized
    """
    pass


class ProducerProduceError(MessagingError):
    """
    A producer could not produce a message
    """
    pass


class ProducerTransactionError(MessagingError):
    """
    A transactional producer or processor could not commit a transaction
    """
    pass



# coding: utf-8
# Copyright (c) Qotto, 2021

from datetime import datetime
from typing import Dict, Any, Union
from uuid import UUID, uuid4

from urnparse import URN8141, InvalidURNFormatError

from .record import Record
from .record_errors import RecordAttributeTypeError, RecordAttributeValueError


class Response(Record):

    def __init__(
        self,
        protocol_version: str,
        schema: str,
        version: str,
        source: str,
        destination: str,
        request_uuid: str,
        uuid: str = None,
        ok: bool = True,
        error_code: int = None,
        error_message: str = None,
        correlation_id: str = None,
        partition_key: str = None,
        date: datetime = None,
        context: Dict[str, Dict[str, Any]] = None,
        data: Dict[str, Any] = None

    ):
        """
        Initialize Record attributes

        :param protocol_version: SemVer version of the Eventy protocol
        :param schema: URN of the record schema
        :param version: SemVer version of the record schema
        :param source: URN of the record emitter
        :param destination: URN of the request destination
        :param request_uuid: UUID v4 of the corresponding request
        :param uuid: UUID v4 of the response. Default: auto-generated
        :param ok:
        :param error_code:
        :param error_message:
        :param correlation_id: Trace / correlation ID for trace_id. Default: auto-generated
        :param partition_key: Partition or database key. Default: None
        :param date: Date of the event. Default: now
        :param context: Keys should be URN of a service using the context
        :param data: Domain data

        :raises RecordAttributeError:
        """
        super().__init__(
            protocol_version,
            schema,
            version,
            source,
            uuid,
            correlation_id,
            partition_key,
            date,
            context,
            data,
        )
        self.destination = destination
        self.request_uuid = request_uuid
        self.ok = ok
        self.error_code = error_code
        self.error_message = error_message

    @property
    def type(self) -> str:
        return 'RESPONSE'

    @property
    def destination(self) -> str:
        return self._destination

    @destination.setter
    def destination(self, destination: str):
        try:
            URN8141.from_string(destination)
        except InvalidURNFormatError:
            raise RecordAttributeValueError('destination', destination, 'Not in URN format')
        self._destination = destination

    @property
    def request_uuid(self) -> str:
        return self._request_uuid

    @request_uuid.setter
    def request_uuid(self, request_uuid: str):
        if request_uuid is None:
            request_uuid = str(uuid4())
        try:
            UUID(request_uuid)
        except TypeError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        except ValueError as e:
            raise RecordAttributeValueError('request_uuid', request_uuid, str(e)) from e
        self._request_uuid = request_uuid

    @property
    def ok(self) -> bool:
        return self._ok

    @ok.setter
    def ok(self, ok: Union[bool, str]):
        if isinstance(ok, bool):
            self._ok = ok
        elif isinstance(ok, str):
            if ok.lower() == 'true':
                self._ok = True
            elif ok.lower() == 'false':
                self._ok = False
            else:
                raise RecordAttributeValueError('ok', bool, ok)
        else:
            raise RecordAttributeTypeError('ok', bool, ok)

    @property
    def error_code(self) -> Union[None, int]:
        return self._error_code

    @error_code.setter
    def error_code(self, error_code):
        # TODO: Check if ok is True? This would impose to update ok before error_code, but why not?
        if error_code is not None and not isinstance(error_code, int):
            raise RecordAttributeTypeError('error_code', 'int or None', error_code)
        self._error_code = error_code

    @property
    def error_message(self) -> Union[None, str]:
        return self._error_message

    @error_message.setter
    def error_message(self, error_message):
        # TODO: Check if ok is True? This would impose to update ok before error_message, but why not?
        if error_message is not None and not isinstance(str, error_message):
            raise RecordAttributeTypeError('error_message', 'str or None', error_message)
        self._error_message = error_message



# coding: utf-8
# Copyright (c) Qotto, 2021

from typing import Iterable

from confluent_kafka import Consumer as KConsumer, Producer as KProducer, Message as KMessage

import logging
from .kafka_errors import (
    KafkaConsumerCreationError,
    KafkaConsumerCommitError,
    KafkaProducerProduceError,
    KafkaProducerCreationError, KafkaConsumerPollError,
)
from ..base import Consumer, Producer
from ...record import Record
from ...serialization import RecordSerializer, ContextRecordSerializer

logger = logging.getLogger(__name__)


class KafkaConsumer(Consumer):
    """Consumer implementation using confluent kafka python client

    See:
        https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#pythonclient-consumer
    """

    def __init__(
        self,
        consumer_config: dict[str, str],
        consumer_topics: list[str],
        consumer_serializer: RecordSerializer,
    ):
        try:
            self._serializer = consumer_serializer
            self._kafka_consumer = KConsumer(consumer_config)
            self._active = False
            self._kafka_consumer.subscribe(
                consumer_topics,
                on_assign=self._on_assign, on_revoke=self._on_revoke, on_lost=self._on_lost
            )
        except Exception as e:
            raise KafkaConsumerCreationError from e
        self._nb_polled = 0

    def _on_assign(self, consumer, partitions):
        logger.debug(f"Partitions assigned {partitions}")

    def _on_revoke(self, consumer, partitions):
        logger.debug(f"Partitions revoked {partitions}")

    def _on_lost(self, consumer, partitions):
        logger.debug(f"Partitions lost {partitions}")

    def poll(self, nb: int = 1, timeout: float = 0, *arg, **kwargs) -> Iterable[Record]:
        try:
            messages: list[KMessage] = self._kafka_consumer.consume(nb, timeout=timeout)
        except Exception as e:
            raise KafkaConsumerPollError from e
        for message in messages:
            if not message:
                logger.debug(f"No message")
                continue

            logger.debug(f"{message.topic()}:{message.partition()}:{message.offset()} Polled kafka message")

            if message.error():
                logger.error(
                    f"{message.topic()}:{message.partition()}:{message.offset()} Message has error: {message.error()}"
                )
                continue

            # PyCharm complains but Message.value() does not have the 'payload' arg, it is the returned value
            # see Confluent API doc: https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#confluent_kafka.Message.value
            # noinspection PyArgumentList
            if not message.value():
                logger.warning(f"{message.topic()}:{message.partition()}:{message.offset()} Message has no value")
                continue

            try:
                # noinspection PyArgumentList
                record = self._serializer.decode(message.value())
                yield record
            except Exception as e:
                logger.error(
                    f"{message.topic()}:{message.partition()}:{message.offset()} Message could not be deserialized",
                    e, )
                continue

    def commit(self):
        try:
            self._kafka_consumer.commit(asynchronous=False)
        except Exception as e:
            # TODO: some exceptions are OK (nothing to commit yet)
            raise KafkaConsumerCommitError from e


class KafkaProducer(Producer):
    def __init__(
        self,
        producer_config: dict[str, str],
        context_serializer: ContextRecordSerializer,
    ):
        self._context_serializer = context_serializer
        self._delivery_status = None
        self._delivery_error = None
        try:
            self._kafka_producer = KProducer(producer_config)
        except Exception as e:
            raise KafkaProducerCreationError from e

    def _on_delivery(self, err, msg):
        if err:
            self._delivery_status = False
            self._delivery_error = err
        else:
            self._delivery_status = True
            self._delivered_msg: KMessage = msg

    def produce_now(self, destination: str, record: Record, *args, **kwargs):
        """Produce and flush"""

        try:
            encoded = self._context_serializer.encode(destination, record)
        except Exception as e:
            raise KafkaProducerProduceError('serialization failed') from e

        try:
            self._kafka_producer.produce(destination, encoded, record.partition_key, on_delivery=self._on_delivery)
            # TODO: should we try to purge() and retry in case of BufferError?
            self._kafka_producer.flush()
        except Exception as e:
            raise KafkaProducerProduceError from e

        if self._delivery_status:
            logger.debug(
                f'Delivered kafka message '
                f'{self._delivered_msg.topic()}:{self._delivered_msg.partition()}:{self._delivered_msg.offset()}'
            )
        else:
            raise KafkaProducerProduceError(f'Record was not delivered because: {self._delivery_error}')



# coding: utf-8
# Copyright (c) Qotto, 2021

from typing import Any, Union


class RecordAttributeError(Exception):
    pass


class RecordAttributeTypeError(RecordAttributeError):
    """Record attribute has wrong type"""

    def __init__(self, attribute: str, expected_type: Union[str, type], value: Any):
        super().__init__(
            f'Record attribute "{attribute}" should be of type "{expected_type}", '
            f'value was "{value}" of type {type(value)}'
        )


class RecordAttributeValueError(RecordAttributeError):
    """Record attribute has wrong value"""

    def __init__(self, attribute: str, value: Any, message: str = None):
        super().__init__(
            f'Record attribute "{attribute}" has incorrect value "{value}" of type {type(value)}'
            f' (cause: {message or "unspecified"})'
        )



# coding=utf-8
# Copyright (c) Qotto, 2021

from os import path

import toml
import pipfile
from pipfile.api import PipfileParser
from setuptools import setup, find_packages

# Read long_description in README.md
# -----------------------------------------------------------------------------
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

# Read dependencies in Pipfile
# -----------------------------------------------------------------------------
parsed_pipfile = PipfileParser(path.join(this_directory, 'Pipfile')).parse()
install_requires = list()
extra_requires = dict()
for pkg, ver in parsed_pipfile.get('default').items():
    if isinstance(ver, str):
        # default
        install_requires.append(f'{pkg}{ver}' if ver != "*" else pkg)
    else:
        # extra
        extra_requires[pkg] = [f'{p}{v}' if v != "*" else p for p, v in ver.items()]
extra_requires['develop'] = [f'{p}{v}' if v != "*" else p for p, v in parsed_pipfile.get('develop').items()]


# Setup
# See: https://packaging.python.org/guides/distributing-packages-using-setuptools/
# -----------------------------------------------------------------------------
setup(
    name="eventy",

    version="3.0.6a2",

    # https://packaging.python.org/specifications/core-metadata/#summary
    description="Qotto/eventy",

    # https://packaging.python.org/specifications/core-metadata/#description-optional
    long_description=long_description,

    long_description_content_type='text/markdown',

    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url="https://gitlab.com/qotto/oss/eventy",

    # This should be your name or the name of the organization which owns the project.
    author="Qotto dev team",

    # This should be a valid email address corresponding to the author listed
    # above.
    author_email="dev@qotto.net",  # Optional

    # For a list of valid classifiers, see https://pypi.org/classifiers/
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        # "Development Status :: 3 - Alpha",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        # "Topic :: Software Development :: Build Tools",
        # Pick your license as you wish
        "License :: OSI Approved :: MIT License",
        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],

    # Note that this is a string of words separated by whitespace, not a list.
    # keywords="sample setuptools development",

    packages=find_packages(exclude=["docs", "tests", "recipes", "examples"]),

    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#python-requires
    python_requires=">=3.7, <4",

    # https://packaging.python.org/en/latest/requirements.html
    install_requires=install_requires,

    # Similar to `install_requires` above, these must be valid existing
    # projects.
    # extras_require={
    #     "dev": [
    #     ]
    # },

    # see https://python-packaging.readthedocs.io/en/latest/dependencies.html#packages-not-on-pypi
    dependency_links=[],

    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    project_urls={  # Optional
        "Qotto": "http://www.qotto.net",
        "Documentation": "https://qotto.gitlab.io/oss/eventy",
        "Source": "https://gitlab.com/qotto/oss/eventy",
        "Bug Reports": "https://gitlab.com/qotto/oss/eventy/-/issues",
    },
)



# coding: utf-8
# Copyright (c) Qotto, 2021

from logging import LogRecord

from ..context import correlation_id_ctx, request_id_ctx


def add_correlation_id_filter(log_record: LogRecord) -> bool:
    log_record.trace_id = correlation_id_ctx.get()  # type: ignore
    return True


def add_request_id_filter(log_record: LogRecord) -> bool:
    log_record.request_id = request_id_ctx.get()  # type: ignore
    return True



# coding: utf-8
# Copyright (c) Qotto, 2021

"""Eventy messaging API

The messaging API itself is backend-agnostic, there is currently only a Kafka backend implemented.
"""

from .base import Producer, Consumer
from .transactional import TransactionalProducer, TransactionalProcessor

__all__ = [
    'Producer',
    'Consumer',
    'TransactionalProducer',
    'TransactionalProcessor',
]



# coding: utf-8
# Copyright (c) Qotto, 2021

from ..messaging_errors import (
    MessagingError,
    ConsumerPollError,
    ProducerTransactionError,
    ConsumerCommitError,
    ProducerProduceError,
    ConsumerCreationError,
    ProducerCreationError,
)


class KafkaError(MessagingError):
    pass


class KafkaConsumerCreationError(KafkaError, ConsumerCreationError):
    pass


class KafkaConsumerPollError(KafkaError, ConsumerPollError):
    pass


class KafkaConsumerCommitError(KafkaError, ConsumerCommitError):
    pass


class KafkaProducerCreationError(KafkaError, ProducerCreationError):
    pass


class KafkaProducerProduceError(KafkaError, ProducerProduceError):
    pass


class KafkaProducerTransactionError(KafkaError, ProducerTransactionError):
    pass


class KafkaTopicCreationError(KafkaError):
    pass


class KafkaTopicDeletionError(KafkaError):
    pass
